<?php
/*
 * Enrober v1.1.0 - PHP theming engine
 *
 * Copyright 2011  Peter Johnson
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/**
 * @package enrober
 * @author Peter Johnson
 * @copyright Copyright (c) 2011, Peter Johnson
 * @version 1.1.0
 * @license http://www.apache.org/licenses/LICENSE-2.0 Apache License, Version 2.0
 * @link http://www.uselesscode.org/php/enrober/
 */

/**
 */

// Try to load customized config
if (file_exists(__DIR__ . '/cfg/config.php')) {
	require_once(__DIR__ . '/cfg/config.php');
} else {
	// if it fails fallback on default
 	require_once(__DIR__ . '/cfg/config.default.php');
}


/**
 * @package enrober
 * @subpackage classes
 */
class Enrober {
	/**
	 * Once an instance of the Enrober class is created using the singleton() method,
	 * it is stored in $instance.
	 * @private
	 */
	private static $instance;

	// Ini to null so default can be set based on config CONSTANTS
	// or overridden by methods or other logic in go()
	private $cache_pages = null; // should caching be used?
	private $cache_valid = null; // is the cache file for the current page still valid

	/**
	 * @var string The currently selected theme
	 */
	private $theme_name = null;

	/**
	 * Array of CSS files to link
	 */
	private $css_stack = array();

	/**
	 * Array of JavaScript files to link
	 */
	private $js_stack = array();

	/**
	 * Array of filters to apply to the body of the page.
	 */
	private $filter_body_stack = array();

	/**
	 * Array of filters to apply to the entire of the page.
	 */
	private $filter_all_stack = array();

	/**
	 * Array of tags that will be processed in the template header.
	 */
	private $tag_stack_head = array();

	/**
	 * Array of tags that will be processed in the template footer.
	 */
	private $tag_stack_foot = array();

	/**
	 * Array of modifiers that will be used in generating the cache file filename.
	 */
	private $modifiers = array();

	/**
	 * String of keywords for a <meta> tag
	 */
	private $keywords = '';

	/**
	 * String of the description for a <meta> tag
	 */
	private $desc = '';

	 /**
	 * The canonical URL for this page
	 */
	private $url = null;

	/**
	 * Is the User Agent mobile, or are we forcing it to
	 * the mobile or non-mobile version.
	 *
	 * Null means go() will auto-detect, if true or false it was manually set via 
	 * force_mobile()
	 */
	private $mobile = null;

	/**
	 * @ignore
	 */
	private function __construct() {
	}

	/**
	 * Creates the $enrober singleton object.
	 * Automatically called when enrober.php is included.
	 * The created object is assigned to the $enrober variable.
	 * Because Enrober is a singleton, further calls to singleton()
	 * will result in a reference to the original object instead
	 * of a new instance.
	 * @returns object
	 */
	public static function singleton() {
		if(!isset(self::$instance)) {
			$c = __CLASS__;
			self::$instance = new $c;
		}

		return self::$instance;
	}

	/**
	 * @ignore
	 */
	public function __clone() {
		trigger_error('Enrober can not be cloned', E_USER_ERROR);
	}

	/**
	 * Returns the complete file system path to the cache file for the current version of the calling script.
	 * {@link set_theme()} must be called before get_cache_file_path() because the 
	 * theme name is used to generate the cache file's filename.
	 * @returns string The path to the cache file
	 */
	public function get_cache_file_path() {
		static $path = null;
		if ($path === null) {
			if ($this->theme_name === null) {
				trigger_error('Can not call $enrober->get_cache_file_path() without calling $enrober->set_theme() first', E_USER_ERROR);
			}
			$path = __DIR__ . '/cache/' . $this->get_cache_filename();
		}

		return $path;
	}

	/**
	 * Override the default theme.
	 *
	 * Can only be called once, subsequent calls will be ignored.  If no $theme is provided, the default
	 * theme is explicitly set as the theme.
	 *
	 * Certain methods (for example {@link get_cache_file_path}) require that the theme be 
	 * set before they are called or they may raise warnings or errors, in these 
	 * cases call set_theme() (with no parameters if you want to use the default 
	 * theme) to avoid errors.
	 *
	 * set_theme() must be called before go(), go() uses the default theme if one has not been
	 * set before go() is called.
	 *
	 * <code>
	 * $enrober->set_theme('pressy'); // Change the them from the default to the 'pressy' theme
	 * $enrober->go();
	 * </code>
	 *
	 * @param string $theme A string containing the name of the theme to wrap the page in.  If not provided the value of the ENROBER_DEFAULT_THEME constant from the config file is used.
	 * @since 1.1.0 Previously the theme could be set as a second parameter to {@link go()}, but it is now required that you use set_theme().
	 */
	public function set_theme($theme = null) {
		if ($this->theme_name === null) {
			if ($theme === null) {
				$theme = ENROBER_DEFAULT_THEME;
			}
			// Theme names can only contain a-z, A-Z, 0-9, underscore and dash to protect
			// against dir traversal in $theme in case users set the theme with
			// a var that can be manipulated by an attacker
			$theme = preg_replace('/[^a-z-_A-Z0-9]/', '', $theme);
		 	if (in_array($theme, $this->get_themes())) {
				$this->theme_name = $theme;
			} else {
				trigger_error('Enrober was unable to find the theme "' . $theme . '"', E_USER_ERROR);
			}
		}
	}

	/**
	 * Returns a list of themes currently available.
	 * @returns array An array containing the names of all of the themes installed in this copy of Enrober
	 * @since 1.1.0
	 */
	public function get_themes() {
		static $themes = null;
		$out = array();

		if ($themes === null) {
			$d = dir(__DIR__ . '/themes/');
			while (($item = $d->read()) !== false) {
				if ($item !== '.' && $item !== '..' && is_dir($d->path . $item)) {
					$out[] = $item;
				}
			}
			$d->close();
		}
		return $out;
	}

	/**
	 * Adds a cache modifier.
	 *
	 * Cache modifiers are used to differentiate between different versions of a page.
	 * This can be useful if you want to cache pages show different information when they
	 * have certain query string parameters, if you want to create different versions of the
	 * page based on the user agent (separate mobile and regular versions for instance), etc.
	 *
	 * @param string $modifier A string that identifies this version of the page. Can contain letters, numbers and underscores.
	 * <code>
	 * <?php
	 * if ($_GET['mode'] === '1') {
	 * 	$enrober->add_cache_modifier('mode1');
	 * }
	 * ?>
	 * </code>
	 */
	public function add_cache_modifier($modifier) {
		if (!in_array($modifier, $this->modifiers)) {
			// Strip any characters that are not alpha-numeric, or dash/underscore
			// since we will be using this as part of a filename, head off dir traversal dangers, etc.
			$modifier = preg_replace('/[^a-zA-Z0-9_]/', '', $modifier);
			$this->modifiers[] = $modifier;
		} else {
			trigger_error($modifier . ' not added to modifier list, already exists in list.', E_USER_WARNING);
		}
	}

	/**
	 * Returns the filename for caching this page
	 * @access private
	 * @returns string
	 */
	private function get_cache_filename() {
		static $name = null;
		if($name === null) {
			$mods = $this->modifiers;
			$enrober_mods = array();

			sort($mods);

			if ($this->theme_name === null) {
				trigger_error('$enrober->get_cache_filename() called before theme_name() set', E_USER_ERROR);
			} else {
				// If it's an XHR request, use a XHR token instead of a theme_name
				if (strtolower($_SERVER['HTTP_X_REQUESTED_WITH']) === 'xmlhttprequest') {
					$name =  'XHR,';
				} else {
					$name = $this->theme_name;
				}
			}

			if ($this->is_mobile() === true) {
				$enrober_mods[] =	'M';
			}

			$name .= ','; // , between theme and mods
			$name .= implode('-', $enrober_mods);
			$name .= ','; // , between Enrober mods and user mods
		 	$name .= implode('-', $mods) . ',' . base64_encode($_SERVER['SCRIPT_NAME']);
		}
		return $name;
	}

	/**
	 * Figure out if there is a cached version of this file/theme/cache-modifier combo.
	 *
	 * {@link set_theme()} must be called before is_cached() because it uses the theme name to figure out the name of the cache file
	 * @returns bool
	 */
	public function is_cached() {
			if ($this->theme_name === null) {
				trigger_error('$enrober->is_cached() called before theme_name() set', E_USER_ERROR);
			}
		return file_exists($this->get_cache_file_path());
	}

	/**
	 * Retrieves a cached copy of this file/theme combo
	 * @returns string The content of the cache file
	 */
	private function get_cache() {
		return file_get_contents($this->get_cache_file_path());
	}

	/**
	 * Manually specify whether or not the file's cache should be invalidated
	 *
	 * @param bool|int $valid - If True assume the cached is still good, if False the file is old and should be regenerated,
	 * if an integer is supplied it is assumed to be a unix timestamp and compared to the cache file's timestamp to determine
	 * the cache file's validity.
	 * <code>
	 * $enrober->manual_cache(true); // cache file is assumed to still be good no 
	 *                               // matter how old it is
	 * $enrober->manual_cache(false); // cache file is assumed to be expired and
	 *                                // must be re-processed
	 * $enrober->manual_cache(mktime(0,0,0,4,1)); // cache file is assumed to be
	 *                                            // expired if the current date is
	 *                                            // after midnight April 1st of the current year
	 * </code>
	 */
	public function manual_cache($valid) {
		// Just set to whatever value we get,
		// true/false/integers are handled in the if($this->is_cached()) section of go()
		if (is_int($valid)) {
			$this->cache_valid = intval($valid);
		} else if (is_bool($valid) || $valid === null) {
			$this->cache_valid = $valid;
		}
	}

 	/**
	 * Figures out the file system path to the current script
	 * @private
	 * @returns string
	 */
	private function get_script_path() {
		$path = realpath($_SERVER['SCRIPT_FILENAME']);
		// realpath can return a funky mixture of / and \ paths
		// make sure the entire path usues the correct DIRECTORY_SEPARATOR
		return preg_replace('/\/|\\\/', DIRECTORY_SEPARATOR, $path); // replaces / or \ with DIRECTORY_SEPARATOR
	}

	/**
	 * Adds <meta> keywords to the page.
	 * Multiple calls append to the list.
	 * @param mixed $keywords What to put in the <meta> keywords tag.  Either a string of comma delimited keywords or an array of keywords
	 */
	public function add_keywords($keywords) {
		if ($this->keywords !== '' && $keywords !== '') {
			$this->keywords .= ',';
		}

		if (is_array($keywords)) {
			$this->keywords .= implode(',', $keywords);
		} else {
			// assume it's a string
			$this->keywords .= $keywords;
		}
	}

	/**
	 * Adds a <meta> description to the page.
	 * Calling add_desc() a second time will replace any previous values.
	 *
	 * @param string $desc What to put in the <meta> description tag
	 */
	public function add_desc($desc) {
		$this->desc = $desc;
	}

	/**
	 * Examines the user agent string to try to determine whether the client is a mobile device.
	 * @returns mixed Contains information about mobile user agents.  If the user agent was not detected as mobile,
	 * it is set to false, otherwise it will be an array with the following keys set to true or false depending
	 * on whether or not the device appears to match the device type.
	 * <ul>
	 * <li>iphone: Device appears to be an iPhone</li>
	 * <li>android: Device appears to be an Android</li>
	 * <li>blackberry: Device appears to be an Black Berry</li>
	 * <li>palm: Device appears to be a Palm device</li>
	 * <li>windows: Device appears to be a Windows Phone or Windows CE device</li>
	 * <li>netfront: Device appears to be running the NetFront browser</li>
	 * <li>safari: Device appears to be running the Safari browser</li>
	 * <li>opera: Device appears to be running the Opera browser</li>
	 * </ul>
	 * @since 1.1.0
	 */
	public function m() {
		static $m = null;

		if ($m === null) {
			$ua = $_SERVER['HTTP_USER_AGENT'];
			if(!preg_match('/MIDP|Mobile|BlackBerry|webOS|^LG[-\/]|NetFront|UP.Browser|Windows (?:CE|Phone)|Palm/', $ua)) {
				$m = false;
			} else {
				$m = array('iphone' => false, 'android' => false, 'blackberry' => false, 'palm' => false, 'netfront' => false, 'safari' => false, 'opera' => false, 'windows' => false);
				// Device
				if(preg_match('/iPhone/', $ua)) {
					$m['iphone'] = true;
				} else if(preg_match('/android/i', $ua)) {
					$m['android'] = true;
				} else if(preg_match('/blackberry/i', $ua)) {
					$m['blackberry'] = true;
				} else if(preg_match('/Palm|webOS/', $ua)) {
					$m['palm'] = true;
				} else if(preg_match('/Windows/', $ua)) {
					$m['windows'] = true;
				}

				// Browsers
				if(preg_match('/netfront/i', $ua)) {
					$m['netfront'] = true;
				} else if(preg_match('/safari/i', $ua)) {
					$m['safari'] = true;
				} else if(preg_match('/opera/i', $ua)) {
					$m['opera'] = true;
				}
			}
		}

		return $m;
	}

	/**
	 * Determines if the page should be served as a mobile page.
	 *
	 * A return value of true does not necessarily mean that the user agent is a mobile device, 
	 * it is possible that {@link force_mobile()} has forced the mobile version.  For a more
	 * definitive answer use {@link m()}.
	 * 
	 * @returns bool
	 * @see m()
	 * @since 1.1.0
	 *
	 */
	public function is_mobile() {
		if (ENROBER_MOBILE_ENABLE) {
			// If we are not forcing mobile/non-mobile...
			if ($this->mobile === null) {
				// Detect mobile browsers
				if ($this->m() === false) {
					return false;
				} else {
					return true;
				}
			}
			// Forced, get value
			return $this->mobile;
		} else {
			// Automatic mobile detection is disabled...
			return false;
		}
	}

	/**
	 * Forces the mobile or non-mobile version of a page to be displayed.
	 *
	 * @param bool|null $force True === force mobile, False === force non-mobile, null === revert to auto-detect
	 * @since 1.1.0
	 */
	public function force_mobile($force) {
		switch ($force) {
		case true:
		case false:
		case null:
			$this->mobile = $force;
			break;
		}
	}

	/**
	 * Invokes Enrober's page wrapping; invoke after calling any of the other Enrober methods you want to use and before page content starts.
	 *
	 * Enrober will automatically call ob_start('ob_gzhandler') on every page it wraps; don't make your own calls to the ob_gzhandler callback
	 * or you will get error messages or garbled gz-encoded text instead of the plain HTML page you intended.
	 *
	 * <code>
	 * <?php
	 * $enrober->go(); // Invoke Enrober without a page title
	 * $enrober->go('Example'); // Invoke Enrober with the page title 'Example'
	 * ?>
	 * </code>
	 *
	 * @param string $title If specified this string will be inserted into a <title> tag
	 */
	public function go($title = '') {


		ob_start("ob_gzhandler");

		if ($this->theme_name === null) {
			$this->set_theme(ENROBER_DEFAULT_THEME);
		}

		$theme = $this->theme_name;

		if (defined('ENROBER_SITE_NAME') && ENROBER_SITE_NAME !== null) {
			$this->add_tag('site_name', ENROBER_SITE_NAME, 1);
		}

		if (defined('ENROBER_TAGLINE') && ENROBER_TAGLINE !== null) {
			$this->add_tag('tagline', ENROBER_TAGLINE, 1);
		}

		$desc = $this->desc;
		$keywords = $this->keywords;

		$js = $this->get_js();
		$css = $this->get_css();
		$url = $this->url;

		// if force_cache() hasn't already overridden the config
		if($this->cache_pages === null) {
			if(!defined('ENROBER_CACHE')) {
				$this->cache_pages = false;
			} else {
				$this->cache_pages = ENROBER_CACHE;
			}
		}

		$theme_fs_path = __DIR__ . "/themes/$theme/";
		$theme_uri_path = ENROBER_URL . "/themes/$theme";

		// If we the client is a mobile UA and the current theme has a mobile version
		if ($this->is_mobile() && file_exists($theme_fs_path . 'm/') ) {
			// change theme paths to the mobile version
			$theme_fs_path .= 'm/';
			$theme_uri_path .= '/m';
		}

		if ($title !== '') {
			$title = '<title>' . (htmlentities($title)) . '</title>';
			$this->add_tag('title', $title, 1, false);
		}

		if ($desc !== '') {
			$this->add_tag('desc', '<meta name="description" content="' . htmlentities($desc) . '">', 1, false);
		}

		if ($keywords !== '') {
			$this->add_tag('keywords', '<meta name="keywords" content="' . htmlentities($keywords) . '">', 1, false);
		}

		if ($this->cache_pages === true) {
			$generated = 'Page cached ' . date(DATE_RFC822);
		} else {
			$generated = 'Page generated ' . date(DATE_RFC822);
		}

		if ($js !== '') {
			$this->add_tag('js', $js, 0, false);
		}

		if ($css !== '') {
			$this->add_tag('css', $css, 1, false);
		}

		if ($url !== null) {
			$this->add_tag('canonical', '<link rel="canonical" href="' . $url . '">', 1, false);
		}

		$this->add_tag('themepath', $theme_uri_path, 0, false);
		$this->add_tag('generated', $generated , 0);

		$head_stack = $this->tag_stack_head;
		$foot_stack = $this->tag_stack_foot;


		// The callback must be an anonymous function so we can use closure
		// to pass variables into the callback
		$enrobe_callback = function ($body) use ($theme_fs_path, $head_stack, $foot_stack) {
			// $body is the body of the page as passed in by ob_start

			// If X-requested-with support is turned on and this is an XHR request
			if (ENROBER_X_REQUESTED_WITH === true && $_SERVER['HTTP_X_REQUESTED_WITH'] === 'XMLHttpRequest') {
				// Set empty header/footer
				$header = '';
				$footer = '';
			} else {
				$theme_config = file_exists($theme_fs_path . 'config.json') ? json_decode(file_get_contents($theme_fs_path . 'config.json')) : array();

				// If head or foot are defined in the config file use that or default
				$head_file = isset($theme_config->head) ? $theme_config->head : 'head.html';
				$foot_file = isset($theme_config->foot) ? $theme_config->foot : 'foot.html';
				// Non-XHR, load header/footer a usual
				$header = file_get_contents($theme_fs_path . $head_file);
				$footer = file_get_contents($theme_fs_path . $foot_file);

				// Process all of the %tags% for the head
				if(count($head_stack) > 0) {
					foreach($head_stack as $tag) {
						$header = preg_replace('/%' . $tag['name'] . '%/', $tag['value'], $header);
					}
				}
				// Process all of the %tags% for the foot
				if(count($foot_stack) > 0) {
					foreach($foot_stack as $tag) {
						$footer = preg_replace('/%' . $tag['name'] . '%/', $tag['value'], $footer);
					}
				}

				// Remove misc add_tag tags that were not processed
				$header = preg_replace('/%[^%\s]*%/', '', $header);
				$footer = preg_replace('/%[^%\s]*%/', '', $footer);
			}

			$content = $header . $body . $footer;

			return $content;
		}; // Anon function, the ; is needed  here!

		$cache_file_path = $this->get_cache_file_path();
		$valid_cache = false;
		if($this->is_cached()) {
			if($this->cache_valid === null) {
				// automatic cache expire based on source file's modified date
				if (filemtime($this->get_script_path()) < filemtime($cache_file_path)) {
					$valid_cache = true;
				}
			} else if(is_int($this->cache_valid)) {
				// expire based on a specified time
				// if the current time is smaller than the expire time, cache is valid
				if ($this->cache_valid > time()) {
					$valid_cache = true;
				}
			} else {
				// cache validity was set manually to true or false
				$valid_cache = $this->cache_valid;
			}
		} else {
			// no cached file, can't be valid!
			$valid_cache = false;
		}

		if ($this->cache_pages && $valid_cache) {
			echo $this->get_cache();
			ob_end_flush(); // we've got our cached file, flush out the ob_gzhandler
			exit;
		} else {

			// If caching is enabled and the current cache file is not valid
			if ($this->cache_pages === true) {
				// cache the page we just generated
				$cache_callback = function ($input) use ($cache_file_path) {
					// Store the page
					file_put_contents($cache_file_path, $input);
					// return the buffer so the page is displayed now too
					return $input;
				}; // Anon function, the ; is needed  here!

				ob_start($cache_callback);
			}

			// Start each filter in the filter_all stack
			// This affects the page content, the header and footer
			for($i = count($this->filter_all_stack)-1; $i >= 0; $i--) { // Apply in reverse so they filter in the right order
				$filter = $this->filter_all_stack[$i];
				ob_start($filter);
			}

			ob_start($enrobe_callback); 

			// Start each filter in the filter_body stack
			// This only affects the page content, not the header or footer...
			for($i = count($this->filter_body_stack)-1; $i >= 0; $i--) { // Apply in reverse so they filter in the right order
				$filter = $this->filter_body_stack[$i];
				ob_start($filter);
			}
		}
		// Function ends, normal page logic takes over; ob_ callbacks end when page ends
	}

	/**
	 * Adds a CSS file to be inserted at the %css% tag.
	 *
	 * <code>
	 * <?php
	 * $enrober->add_css('extra.css'); // Linked as a regular stylesheet
	 * // Linked as a print-only stylesheet
	 * $enrober->add_css('print.css', array('media' => 'print');
	 * // Linked as an alternative stylesheet
	 * $enrober->add_css('alternate.css', array('rel' => 'alt', 'title' => 'Alternate layout');
	 * ?>
	 * </code>
	 *
	 * @param string $filepath The url to the css file to be included.  If the css file is in the same directory as the calling script only the filename needs to be specified.
	 * @param array $config An array that allows you to add additional optional attributes to the link tag.
	 * These attributes are:
	 * <ul>
	 * <li>'media': Which media types the stylesheet should be applied to, 'all' by default.</li>
	 * <li>'rel': The rel attribute of the <link> tag, 'stylesheet' by default can also be set to 'alternate stylesheet' to specify an alternate stylesheet.  If set to 'alt' it will automatically be expanded to 'alternate stylesheet'.</li>
	 * <li>'title': The title attribute for the link, most browsers that support alternative stylesheets will use this as a label in the stylesheet switching interface.</li>
	 * </ul>
	 *
	 */
	public function add_css($filepath, $config = null) {

		if (!is_array($config)) {
			$config = array();
		}

		if (array_key_exists('rel', $config)) {
			// Allow $rel to be set to 'alt' as a shortcut for 'alternate stylesheet'
			if ($config['rel'] === 'alt') {
				$config['rel'] = 'alternate stylesheet';
			}
		} else {
			$config['rel'] = 'stylesheet';
		}

		if (!array_key_exists('media', $config)) {
			$config['media'] = 'all';
		}

		if (!array_key_exists('title', $config)) {
			$config['title'] = 'Stylesheet #' . count($this->css_stack) + 1;
		}

		$new_item = array('file' => $filepath, 'media' => $config['media'], 'rel' => $config['rel'], 'title' => $config['title']);

		/* Don't add if it's already in the stack */
		if (!in_array($new_item,$this->css_stack,true)) {
			$this->css_stack[] = $new_item;
		} else {
			trigger_error($filepath . ' not added to CSS stack, already exists in stack.', E_USER_WARNING);
		}
	}

	/**
	 * Adds a JavaScript file to be linked via the %js% tag.
	 * 
	 * @param string $filepath The url of the JavaScript file to add absolute or relative to the calling script.
	 *
	 * <code>
	 * <?php
	 * $enrober->add_js('http://cdn.example.com/1.4.4/jquery.min.js'); // Import jQuery from a CDN
	 * $enrober->add_js('blink.js'); // Add a JavaScript file that is in the same
	 *                               // directory as the page using it.
	 * ?>
	 * </code>
	 */
	public function add_js($filepath) {
		/* Don't add if it's already in the stack */
		if (!in_array($filepath, $this->js_stack)) {
			$this->js_stack[] = $filepath;
		} else {
			trigger_error($filepath . ' not added to JS stack, already exists in stack.', E_USER_WARNING);
		}
	}

	/**
	 * Adds a content filter to be applied with ob_start()
	 *
	 * @param string $filter A string with the name of the filter function to use
	 * @param bool $filter_everything Should only the body be filtered or the entire generated page? (By default only the body of the page is filtered)
	 *
	 * <code>
	 * <?php
	 *
	 * // Apply the HTML Tidy callback to the body of the
	 * // page to fix any broken markup (Requires the Tidy
	 * // extension for PHP). We are only applying it to
	 * // the body because the header/footer should already
	 * // contain valid markup.
	 * $enrober->add_filter('ob_tidyhandler');
	 *
	 * // This filter function collapses any whitespace
	 * // between HTML tags down to a single space
	 * function collapse_whitespace($in) {
	 * 	return preg_replace('/>\s+</','> <', $in); 
	 * }
	 *
	 * // add our custom filter, filtering everything
	 * // including the header and footer
	 * $enrober->add_filter('collapse_whitespace', true);
	 * ?>
	 * </code>
	 */
	public function add_filter($filter, $filter_everything = false) {
		if($filter_everything) {
			$this->filter_all_stack[] = $filter;
		} else {
			$this->filter_body_stack[] = $filter;
		}
	}

	/**
	 * Adds a new theme tag.
	 *
	 *
	 * Theme tags allow you to specify where things Enrober generates are injected into the theme. There are a number of built in tags:
	 * <ul>
	 * <li>%title% - The page's <title> tag</li>
	 * <li>%desc% - The page's description <meta> tag</li>
	 * <li>%keywords% - The page's keywords <meta> tag</li>
	 * <li>%themepath% - The URI for the directory that the current theme is stored in, useful for including the theme's CSS file.</li>
	 * <li>%css% - Any <link> tags for CSS files added via $enrober->add_css()</li>
	 * <li>%js% - Any <script> tags for JavaScript files added via $enrober->add_js()</li>
	 * <li>%generated% - Inserts a "Generated on" or "Cached on" message depending on whether or not caching is enabled on the current page.</li>
	 * <li>%canonical% - Where in the head to insert the Canonical URL if one was set with $enrober->url()</li>
	 * <li>%site_name% - The value for ENROBER_SITE_NAME specified in enrober-config.php</li>
	 * <li>%tagline% - The value for ENROBER_TAGLINE specified in enrober-config.php</li>
	 * </ul>
	 * 
	 * If you add your own tags to your templates you can then set their value for any particular page with add_tag().
	 * 
	 *
	 * <code>
	 * <?php
	 * $enrober->add_tag('foo', 'bar'); // Adds the text "bar" anywhere in the 
	 *                                  // header or footer that "%foo%" appears
	 * $enrober->add_tag('foo', 'bar', 0); // Same as previous
	 * $enrober->add_tag('foo', 'bar', 1); // Adds the text "bar" only in the header
	 * $enrober->add_tag('foo', 'bar', 2); // Adds the text "bar" only in the footer
	 *
	 * // By default anything added to a theme tag is run through htmlentities()
	 * // to prevent XSS attacks. If you need to use HTML you can set the $sanitize
	 * // parameter to false but if you are including any user generated content
	 * // in it you will need to sanitize it yourself!
	 * $enrober->add_tag('foo', '<em>bar</em>'); // Adds the text "bar", wrapped
	 *                                           // in a text representation of em tags
	 * $enrober->add_tag('foo', '<em>bar</em>', 0, true); // Same as previous
	 * $enrober->add_tag('foo', '<em>bar</em>', 0, false); // Adds the text "bar",
	 *                                                     // wrapped in em tags
	 * ?>
	 * </code>
	 *
	 * @param string $name The name of the theme tag to define (sans %%)
	 * @param string $value The value to insert at the new theme tag
	 * @param integer $location Where to process the tag.  0 = Header and Footer, 1 = Header only, 2 = Footer only
	 * @param bool $sanitize If true, the $value is run through htmlentities() before inclusion in the page. Set to false if you need to include HTML, but you will need to sanitize any user input you include yourself!
	 */
	public function add_tag($name, $value, $location = 0, $sanitize = true) {

		if (preg_match('/^%|%$/', $name)) {
			$name = preg_replace('/^%|%$/', '', $name);
			trigger_error('Added theme tag "' . $name . '" with % signs', E_USER_WARNING);
		}

		if ($sanitize !== false) {
			$value = htmlentities($value);
		}

		$new_item = array('name' => $name, 'value' => $value);

		if ($location === 0 || $location === 1) {
			$this->tag_stack_head[] = $new_item;
		}
		if ($location === 0 || $location === 2) {
			$this->tag_stack_foot[] = $new_item;
		}
	}

	/**
	 * Sets the canonical URL for the page to be inserted at the %canonical% tag.
	 *
	 * Google's webmaster central has more information about {@link http://www.google.com/support/webmasters/bin/answer.py?answer=139066 canonical URLs}.
	 * @param string $url The url to set as the canonical url.
	 */
	function url($url) {
		$this->url = $url;
	}

	/**
	 * Manually turn on/off caching, overriding the ENROBER_CACHE constant from the config file.
	 *
	 * When caching is turned on with force_cache(), normal caching rules apply.  To guarantee that 
	 * the cached version of a page gets returned when one exists use {@link manual_cache()} set to true.
	 *
	 * @param bool $cache Should caching be enabled for this page?  True = Cache, False = Don't cache
	 */
	public function force_cache($cache) {
		if($cache === true) {
			$this->cache_pages = true;
		} else {
			$this->cache_pages = false;
		}
	}

	/**
	 * Returns a list of <link> tags for each CSS file in the css stack.
	 * @returns string
	 */
	private function get_css() {
		$out = '';
		foreach($this->css_stack as $stylesheet) {
			$out .= '<link href="' . $stylesheet['file'] . '" rel="' . $stylesheet['rel'] . '" type="text/css" media="' . $stylesheet['media'] . '" title="' . $stylesheet['title'] . '">';
		}
		return $out;
	}

	/**
	 * Returns a string containing <script> tags to add to the document via the %js% tag
	 * @returns string
	 */
	private function get_js() {
		$out = '';
		foreach($this->js_stack as $js_file) {
			$out .= '<script src="' . $js_file . '" type="text/javascript"></script>';
		}
		return $out;
	}
}
// Create our $enrober object for scripts to invoke
$enrober = Enrober::singleton();
?>
